from backend.models import UrlRule, FedUrlRule
from backend.models.url_rule import UrlRuleMatchingType, UrlRuleMode
import backend.controllers as ctrls
from backend.schemas import UrlRuleSchema, NotProvided
from werkzeug.exceptions import NotFound, BadRequest
from marshmallow.exceptions import ValidationError
import re
from sqlalchemy import func
from typing import List


class UrlRuleController(ctrls.BaseController):
    def get_all(self) -> List[UrlRule]:
        return UrlRule.query.all()

    def get_all_for_view(self, view_id: int) -> List[UrlRule]:
        return UrlRule.query.filter_by(view_id=view_id).all()

    def count_rules_for_view(self, view_id) -> int:
        count = self.db.session.query(func.count(UrlRule.id)).filter(UrlRule.view_id == view_id).first()
        if len(count) > 0:
            return count[0]
        else:
            return 0

    def get_by_id(self, url_rule_id: int, ignore_if_not_found: bool = False) -> UrlRule:
        if url_rule_id is None:
            if not ignore_if_not_found:
                raise ValidationError(message={"url_rule_id": ["Field may not be null."]})
            return None
        url_rule = UrlRule.query.get(url_rule_id)
        if url_rule is None and not ignore_if_not_found:
            raise NotFound("URL Rule not found.")
        return url_rule

    def create(
        self,
        view_id: int,
        mode: UrlRuleMode,
        matching_type: UrlRuleMatchingType,
        priority: int,
        value: str,
        css_selectors: List[str] = NotProvided,
    ) -> UrlRule:
        _ = ctrls.view_ctrl.get_by_id(view_id)

        new_url_rule = UrlRuleSchema().load(
            {
                "view_id": view_id,
                "mode": mode,
                "matching_type": matching_type,
                "priority": priority,
                "value": value,
                "css_selectors": css_selectors,
            },
            session=self.db.session,
        )

        self.db.session.add(new_url_rule)
        self.db.session.commit()

        return self.get_by_id(new_url_rule.id)

    def create_many(self, view_id: int, url_rules: List[dict]) -> List[UrlRule]:
        new_url_rules = []
        for url_rule in url_rules:
            new_url_rule = self.create(view_id=view_id, **url_rule)
            new_url_rules.append(new_url_rule)
        return new_url_rules

    def update(
        self,
        url_rule_id: int,
        mode: UrlRuleMode = NotProvided,
        matching_type: UrlRuleMatchingType = NotProvided,
        priority: int = NotProvided,
        value: str = NotProvided,
        css_selectors: List[str] = NotProvided,
    ) -> UrlRule:
        url_rule = self.get_by_id(url_rule_id)

        new_url_rule = UrlRuleSchema().load(
            {
                "mode": mode,
                "matching_type": matching_type,
                "priority": priority,
                "value": value,
                "css_selectors": css_selectors,
            },
            instance=url_rule,
            session=self.db.session,
            partial=("id", "view_id"),
        )

        self.db.session.add(new_url_rule)
        self.db.session.commit()

        return self.get_by_id(url_rule_id)

    def delete(self, url_rule_id: int) -> bool:
        url_rule_to_delete = self.get_by_id(url_rule_id)

        self.db.session.delete(url_rule_to_delete)
        self.db.session.commit()

        is_url_rule_deleted = self.get_by_id(url_rule_id, ignore_if_not_found=True) == None
        if not is_url_rule_deleted:
            raise BadRequest("Could not delete url rule.")

        return True

    def export(self, url_rule_id: int):
        url_rule = self.get_by_id(url_rule_id)

        fields_to_export = ["mode", "matching_type", "priority", "value", "css_selectors"]
        return UrlRuleSchema(only=tuple(fields_to_export)).dump(url_rule)

    def convert_contains_to_regex(self, s: str) -> str:
        escaped_s = re.escape(s)
        return f"^.*{escaped_s}.*$"

    def convert_startswith_to_regex(self, s: str) -> str:
        escaped_s = re.escape(s)
        return f"^{escaped_s}.*$"

    def convert_equals_to_regex(self, s: str) -> str:
        escaped_s = re.escape(s)
        return f"^{escaped_s}$"

    def convert_matches_to_regex(self, s: str) -> str:
        return f"^{s}$"

    def url_rule_to_regex(self, url_rule_id: int) -> str:
        url_rule = self.get_by_id(url_rule_id)
        if url_rule.matching_type == UrlRuleMatchingType.CONTAINS:
            url_rule_regex = self.convert_contains_to_regex(url_rule.value)
        elif url_rule.matching_type == UrlRuleMatchingType.STARTS_WITH:
            url_rule_regex = self.convert_startswith_to_regex(url_rule.value)
        elif url_rule.matching_type == UrlRuleMatchingType.EQUALS:
            url_rule_regex = self.convert_equals_to_regex(url_rule.value)
        elif url_rule.matching_type == UrlRuleMatchingType.MATCHES:
            url_rule_regex = self.convert_matches_to_regex(url_rule.value)
        else:
            raise BadRequest("Wrong URL rule type")
        return url_rule_regex

    def feed_url_rule(self, url_rule_id: int) -> FedUrlRule:
        url_rule = self.get_by_id(url_rule_id)
        url_rule_regex = self.url_rule_to_regex(url_rule_id)
        return FedUrlRule(
            url_rule_id=url_rule_id,
            mode=url_rule.mode,
            regex=url_rule_regex,
            priority=url_rule.priority,
            css_selectors=url_rule.css_selectors,
        )

    def update_priority(self, url_rule_id: int, new_index: int, old_index: int) -> bool:
        url_rule = self.get_by_id(url_rule_id)
        view_rules_nb = self.count_rules_for_view(url_rule.view_id)

        new_priority = view_rules_nb - new_index
        old_priority = view_rules_nb - old_index

        inf_priority, sup_priority = list(sorted([new_priority, old_priority]))
        incr = +1 if new_priority < old_priority else -1

        url_rules_to_update = UrlRule.query.filter(
            UrlRule.priority >= inf_priority, UrlRule.priority <= sup_priority
        ).all()

        for r in url_rules_to_update:
            r.priority = r.priority + incr

        url_rule.priority = new_priority

        self.db.session.add_all(url_rules_to_update)
        self.db.session.commit()

        return True
