import flask_migrate
import os
from flask import Flask
from flask_migrate import Migrate
from backend.utils.dataiku_api import dataiku_api
from backend.utils.misc import get_webapp_id, extend_dataiku_variables

from backend.routes import api_bp
from backend.routes import main_bp

from flask_sqlalchemy import SQLAlchemy

from flask_cors import CORS

from werkzeug.exceptions import HTTPException
from marshmallow import ValidationError

from flask import jsonify, Flask

import traceback
import json


def upgrade_db():
    current_directory = os.path.dirname(os.path.abspath(__file__))
    migration_path = os.path.join(current_directory, "../migrations")
    flask_migrate.upgrade(directory=migration_path)


def set_db_prefix_env():
    config = dataiku_api.webapp_config
    preset_config = config["connections"]

    webapp_id = get_webapp_id()
    db_prefix = extend_dataiku_variables(preset_config.get("db_prefix", ""), webapp_id)
    os.environ["DB_PREFIX"] = db_prefix


def get_conn_url():
    config = dataiku_api.webapp_config
    preset_config = config["connections"]
    storing_method = preset_config["storing_method"]

    webapp_id = get_webapp_id()

    if storing_method == "postgres":
        db_conn_name = preset_config["postgres_connection"]
        db_conn = dataiku_api.get_connection_info(db_conn_name)["params"]
        db_conn_url = (
            f"postgresql://{db_conn['user']}:{db_conn['password']}@{db_conn['host']}:{db_conn['port']}/{db_conn['db']}"
        )
    elif storing_method == "mysql":
        db_conn_name = preset_config["mysql_connection"]
        db_conn = dataiku_api.get_connection_info(db_conn_name)["params"]
        db_conn_url = f"mysql+pymysql://{db_conn['user']}:{db_conn['password']}@{db_conn['host']}:{db_conn['port']}/{db_conn['db']}"
    else:
        db_dir = os.path.abspath(os.path.expanduser(preset_config["fs_path"]))

        db_dir = extend_dataiku_variables(db_dir, webapp_id)
        if not os.path.exists(db_dir):
            os.makedirs(db_dir)
        db_conn_url = f"sqlite:///{os.path.join(db_dir, 'advisor.db')}"
    return db_conn_url


def setup_app_for_db(app_to_setup: Flask, db: SQLAlchemy):
    db_conn_url = get_conn_url()

    app_to_setup.config["SQLALCHEMY_DATABASE_URI"] = db_conn_url
    app_to_setup.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    _ = Migrate(app_to_setup, db)

    db.init_app(app_to_setup)


def run_migration_upgrade_func(client_setup=None):
    from backend.models import db

    migration_app = Flask("migration_app")
    if client_setup:
        dataiku_api.setup(**client_setup)
    setup_app_for_db(migration_app, db)
    _ = Migrate(migration_app, db)

    with migration_app.app_context():
        upgrade_db()

    return

def run_create_app(app: Flask, db: SQLAlchemy):

    setup_app_for_db(app, db)

    app.register_error_handler(HTTPException, handle_exception)
    app.register_error_handler(ValidationError, handle_validation_error)

    CORS(app)

    app.config["WEB_APP_ID"] = get_webapp_id()
    print(get_webapp_id())
    app.register_blueprint(api_bp)
    app.register_blueprint(main_bp)

    return app


def handle_exception(e):
    """Return JSON instead of HTML for HTTP errors."""
    response = e.get_response()
    data = {
        "code": e.code,
        "name": e.name,
        "type": e.name,
        "description": e.description,
        "details": getattr(e, "details", {}),
    }
    if os.getenv("FLASK_ENV") == "development":
        data["trace"] = traceback.format_exc()
    response.data = json.dumps(data)

    response.content_type = "application/json"
    return response


def handle_validation_error(e):
    error_fields = None
    generic_description = "Some fields are malformed: {error_fields}. See details field for more information."
    if type(e.messages) is dict:
        error_fields = ", ".join(e.messages.keys())
        description = generic_description.format(error_fields=error_fields)
        details = e.messages
    elif type(e.messages) is list and len(e.messages) == 1:
        description = e.messages[0]
        details = e.kwargs.get("details", {})
    else:
        description = "Validation Error. See details field for more information."
        details = e.messages

    data = {
        "code": 400,
        "name": "Bad Request",
        "type": "Validation Error",
        "description": description,
        "details": details,
    }
    if os.getenv("FLASK_ENV") == "development":
        data["trace"] = traceback.format_exc()
    response = jsonify(data)

    response.status = 400
    return response
