import logging
import re
import timeit
from functools import wraps
from marshmallow import ValidationError
import json
import sqlalchemy
from sqlalchemy.types import (
    String,
    TypeDecorator,
)
from sqlalchemy.dialects import postgresql
import numpy as np
import pandas as pd
import os

from datetime import datetime
import logging

logger = logging.getLogger(__name__)


class NotProvidedClass(object):
    def __bool__(self):
        return False


NotProvided = NotProvidedClass()


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, pd.Timestamp):
            return str(obj)
        if isinstance(obj, datetime):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        logger.info(f"Starting {f.__name__}...")
        start_time = timeit.default_timer()
        result = f(*args, **kw)
        delta = timeit.default_timer() - start_time
        logger.info(f"{f.__name__}: {delta * 1000} ms")
        return result

    return wrap


class timing_ctx(object):
    def __init__(self, *args) -> None:
        self.args = args

    def __enter__(self):
        self.t = timeit.default_timer()
        return self

    def __exit__(self, type, value, traceback):
        delta = timeit.default_timer() - self.t
        args = ",".join([*self.args])
        logger.info(f"{args}: {delta * 1000} ms")


def parse_req(request, required_fields=[], optional_fields=[], useless_fields=[], allow_additional_fields=False):
    def is_required(field):
        return field in required_fields

    def is_useless(field):
        return field in useless_fields

    req_json = request.get_json(force=True)
    data = {}
    missing_fields = []
    for field in required_fields + optional_fields + useless_fields:
        if isinstance(field, dict):
            field_from = list(field.keys())[0]
            field_to = list(field.values())[0]
        else:
            field_from = field
            field_to = field
        if field_from not in req_json and is_required(field):
            missing_fields.append(field_from)
        else:
            field_value = req_json.pop(field_from, NotProvided)
            if not is_useless(field):
                data[field_to] = field_value

    if len(missing_fields) > 0:
        raise ValidationError(f'Missing mandatory field(s): {", ".join(missing_fields)}.')

    if not allow_additional_fields:
        if len(req_json) > 0:
            raise ValidationError(f'Additional field(s) not allowed: {", ".join(list(req_json.keys()))}.')
    return data


def flatten(el):
    if not isinstance(el, list):
        return [str(el)]
    if el == []:
        return []
    l = []
    for sub_el in el:
        flatten_el = flatten(sub_el)
        if flatten_el is not None:
            l += flatten_el
    return l


def are_instances_equal(instance1, instance2, fields=None):
    if fields is None:
        fields = instance1.__dict__.keys()
    for field in fields:
        if getattr(instance1, field) != getattr(instance2, field):
            return False
    return True


def is_base64_encoded(s):
    pattern = re.compile("^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$")
    if (s is None) or (len(s) == 0):
        return False
    return pattern.match(s)


def validateURL(value):
    regex = r"^[a-z]+://(?P<host>[^/:]+)(?P<port>:[0-9]+)?(?P<path>\/.*)?$"
    return re.match(regex, value)


class ArrayType(TypeDecorator):
    """Sqlite-like does not support arrays.
    Let's use a custom type decorator.

    See http://docs.sqlalchemy.org/en/latest/core/types.html#sqlalchemy.types.TypeDecorator
    """

    impl = String(1000)

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        return json.loads(value)

    def copy(self):
        return ArrayType(self.impl.length)


def get_webapp_id():
    default_webapp_id = "UNKNOWN"
    webapp_id = default_webapp_id
    try:
        current_dir = os.path.abspath(os.getcwd())
        current_dir_splitted = current_dir.split("/")
        if len(current_dir_splitted) >= 1:
            webapp_id = current_dir_splitted[-1]
        else:
            logging.error(f"Could not find Webapp ID (current_dir is {current_dir}). Assigning to UNKNOWN")
    except Exception as err:
        logger.error("Something went wront getting the webapp ID: ", err)
    return webapp_id


def extend_dataiku_variables(s: str, webapp_id: int) -> str:
    replace_variables = dict({"webAppId": webapp_id})
    new_s = s
    for key, value in replace_variables.items():
        new_s = new_s.replace("$_{" + key + "}", value)
    return new_s


def unescape_value(value: str) -> str:
    def linebreak_repl(match):
        if match.group("linebreak"):
            return "  \n"
        elif match.group("bksl"):
            return f'{match.group("bksl")}{match.group("tag")}'
        else:
            return ""

    r = r" *(?:(?:(?:(?P<bksl>(?:\\)+)(?P=bksl))(?P<tag>[a-zA-Z]))|(?P<linebreak>(\\n|\n)))"
    new_val = re.sub(r, linebreak_repl, value)
    return new_val


logger = logging.getLogger(__name__)


class timeLogger:
    def __init__(self) -> None:
        self._start_time = datetime.now()
        self._last_call = datetime.now()
        self._counter = 0

    def time_stamp(self, message: str = None) -> None:
        current_time = datetime.now()
        time_ellapsed_from_last_call = (current_time - self._last_call).total_seconds()
        time_ellapsed_from_start = (current_time - self._start_time).total_seconds()
        self._last_call = current_time
        message = message if message != None else self._counter
        self._counter += 1
        logger.info(
            f"time loger - {message}:\nfrom start:{time_ellapsed_from_start}\nfrom prev:{time_ellapsed_from_last_call}"
        )
        logger.info("=======================")


def full_table_name(table_name):
    db_prefix = os.getenv("DB_PREFIX", "")
    return f"{db_prefix}{table_name}"


def build_type(type_enum, conn, current_dialect):
    type_name = type_enum.__name__.lower()
    if current_dialect == "postgresql":
        new_type = postgresql.ENUM(type_enum, name=type_name, create_type=False)
        new_type.create(conn, checkfirst=True)
    else:
        new_type = sqlalchemy.Enum(type_enum, name=type_name)
    return new_type
